import logging

import relaunch_instance as ri
import relaunch_instance.post as post
import relaunch_instance.settings as settings
import relaunch_instance.util as util


def detach_secondary_nics(instance, rollback, client, resource):
    log = logging.getLogger(__name__)

    rollback["secondary_nics"] = []
    if len(instance.network_interfaces) > 1:
        log.info(f"Instance: {instance.instance_id} has multiple nics")
        for nic in instance.network_interfaces_attribute:
            if nic["Attachment"]["DeviceIndex"] == 0:
                continue
            detached_nic = util.instance.detach_nic(nic, client, resource)
            if not detached_nic:
                post.rollback(instance, resource, client, rollback)
                return False, None
            rollback["secondary_nics"].append(detached_nic)

    return rollback, rollback["secondary_nics"]


def detach_eip(instance, rollback, client, resource):
    log = logging.getLogger(__name__)

    eip = None
    instance.update()
    if instance._i.public_ip_address:
        log.info(f"Instance: {instance.instance_id} has static EIP address")
        eip = util.instance.disassociate_public_ip(instance, client, resource)
        if not eip:
            post.rollback(instance, resource, client, rollback)
            return False, None
        rollback["eip"] = eip

    return rollback, eip


def replace_instance(instance, i_settings, region, client, resource):
    """
    Remove secondary nics
    Remove EIP
    Terminate Old instance
    Launch new instance
    Reassocaite EIP
    """
    log = logging.getLogger(__name__)
    rollback = {}

    try:
        rollback, nics = detach_secondary_nics(instance, rollback, client, resource)
        if not rollback:
            return 1

        rollback, eip = detach_eip(instance, rollback, client, resource)
        if not rollback:
            return 1

        if not util.instance.dry_run(client, i_settings):
            post.rollback(instance, resource, client, rollback)
            return False

        # TODO if instance has stop protection it won't be enabled again. This needs to be fixed
        if not util.instance.terminate(instance):
            post.rollback(instance, resource, client, rollback)
            return False

        # Reset rollback because instance has been terminated
        rollback = {}

        new_instance = run_instance(instance, i_settings, region, client, resource)
        if not new_instance:
            settings.dump(i_settings)
            return False

        if eip:
            if nics:
                primary = util.instance.get_primary_nic(
                    new_instance.network_interfaces_attribute
                )
                util.eip.associate(eip, primary["NetworkInterfaceId"])
            else:
                util.eip.associate(eip, new_instance.instance_id)

        return new_instance
    except Exception as e:
        log.error(f"Unexpected Failure: {e}")
        post.rollback(instance, resource, client, rollback)
        return False


def launch_new_instance(instance, i_settings, region, client, resource):
    """
    Remove secondary nics
    Remove EIP
    Rename old instance
    Launch new instance
    Reassociate EIP
    """
    log = logging.getLogger(__name__)
    rollback = {}

    try:
        rollback, nics = detach_secondary_nics(instance, rollback, client, resource)
        if not rollback:
            return 1

        rollback, eip = detach_eip(instance, rollback, client, resource)
        if not rollback:
            return 1

        tags = util.get_tag_dict(instance.tags)
        if not util.instance.rename(instance, tags["Name"] + "-OLD"):
            post.rollback(instance, resource, client, rollback)
        rollback["name"] = tags["Name"]

        new_instance = run_instance(instance, i_settings, region, client, resource)
        if not new_instance:
            settings.dump(i_settings)
            post.rollback(instance, resource, client, rollback)
            return False

        if eip:
            if nics:
                primary = util.instance.get_primary_nic(
                    new_instance.network_interfaces_attribute
                )
                util.eip.associate(eip, primary["NetworkInterfaceId"])
            else:
                util.eip.associate(eip, new_instance.instance_id)

        return new_instance
    except Exception as e:
        log.error(f"Unexpected Failure: {e}")
        post.rollback(instance, resource, client, rollback)
        return False


def run_instance(instance, i_settings, region, client, resource):
    log = logging.getLogger(__name__)
    try:
        new_instance = util.instance.run(client, resource, i_settings)
        if not new_instance:
            return False

        util.instance.copy_volume_tags(instance, new_instance, region, resource)

        return new_instance
    except Exception as e:
        log.error(f"Unexpected Failure: {e}")
        return False


def relaunch(
    instance_id,
    region,
    client,
    resource,
    inference_client,
    ami_id=None,
    instance_type=None,
    private_ip=None,
    root_volume_size=None,
    encrypt_root_volume=False,
    kms_key_id=None,
    enable_hibernation=False,
    settings_input=None,
    dump_settings=False,
    settings_output=ri.SETTINGS_FILE,
    launch_template=False,
    log_level="INFO",
    yes=False,
):
    log = logging.getLogger(__name__)
    logging.basicConfig(level=log_level)

    if yes:
        ri.YES = True

    log.info("Starting")
    instance = util.instance.get(instance_id, client, resource)
    if not instance:
        log.error(f"Instance: {instance_id} not found in {region}")
        return 2

    if "Name" not in util.get_tag_dict(instance.tags):
        log.warning("Please tag your instance with a Name tag")
        return 2

    overrides = settings.Overrides(
        ami_id,
        instance_type,
        private_ip,
        root_volume_size,
        encrypt_root_volume,
        kms_key_id,
        enable_hibernation,
    )
    i_settings, rollback = settings.get(
        instance,
        region,
        overrides,
        settings_input,
        client,
        resource,
        inference_client,
    )
    if type(i_settings) == int:
        return i_settings

    if launch_template:
        launch_template_data = i_settings
        launch_template_data.pop("MaxCount")
        launch_template_data.pop("MinCount")
        name = util.get_tag_dict(instance.tags)["Name"] + "-RELAUNCH_TEMPLATE"
        client.create_launch_template(
            LaunchTemplateName=name, LaunchTemplateData=launch_template_data
        )
        log.info("Launch Template Generated")
        return 0

    if dump_settings:
        settings.dump(i_settings, settings_output)
        if rollback:
            rollback["ami"] = None
            post.rollback(instance, resource, client, rollback)
            log.info("Finished")
        return 0

    # Settings have been generated or dumped.
    # Instance is stopped (only if settings were generated)
    # EIP is still associated
    # AMI has been created

    primary_nic = util.instance.get_primary_nic(i_settings["NetworkInterfaces"])
    existing_primary_nic = util.instance.get_primary_nic(
        instance.network_interfaces_attribute
    )
    if util.vpc.ip_conflict(
        primary_nic, existing_primary_nic["NetworkInterfaceId"], client
    ):
        return 1

    if util.instance.ip_conflict([primary_nic, existing_primary_nic]):
        log.warning("Old instance must be deleted to move private IP")
        return replace_instance(instance, i_settings, region, client, resource)

    return launch_new_instance(instance, i_settings, region, client, resource)
