import json
import logging
from dataclasses import dataclass

import relaunch_instance as ri
import relaunch_instance.hibernate as hibernate
import relaunch_instance.post as post
import relaunch_instance.util as util


@dataclass
class Overrides:
    """
    Override instance settings.
    """

    ami_id: str = None
    instance_type: str = None
    private_ip: str = None
    root_volume_size: str = None
    encrypt_root_volume: bool = False
    kms_key_id: str = None
    enable_hibernation: str = None


def get(
    instance,
    region,
    overrides,
    settings_input_file,
    client,
    resource,
    inference_client,
):
    log = logging.getLogger(__name__)

    if overrides.enable_hibernation:
        res = hibernate.validate_instance(
            instance, overrides.instance_type, region, client
        )
        if res != 0:
            return res, None

    i_settings, rollback = None, None
    if settings_input_file:
        log.info("Using settings from file")
        return load(settings_input_file), None

    i_settings, rollback = generate(
        instance, overrides, client, resource, inference_client
    )

    if not i_settings:
        return 1, None

    return i_settings, rollback


def generate(instance, overrides, client, resource, inference_client):
    log = logging.getLogger(__name__)
    rollback = {}

    try:
        log.info("Generating instance settings")

        rollback["is_running"] = True if instance.state["Name"] != "stopped" else False
        if not util.instance.stop(instance):
            return None, None

        ami = None
        if overrides.ami_id:
            ami = util.ami.get(overrides.ami_id, client, resource)
        else:
            tags = util.get_tag_dict(instance.tags)
            ami_name = tags["Name"] + "-RELAUNCH"
            ami = util.ami.create(instance, ami_name, client, resource, tags=tags)
            rollback["ami"] = ami

        if not ami:
            post.rollback(instance, resource, client, rollback)
            return None, None

        i_settings = prepare_kwargs(instance, ami, overrides, client, inference_client)
        return i_settings, rollback

    except Exception as e:
        log.error(f"Unexpected Failure while getting instance settings: {e}")
        post.rollback(instance, resource, client, rollback)
        return None, None


def prepare_kwargs(instance, ami, overrides, client, inference_client):
    """
    Generate instance settings.
    Currently only generates tags for instance.
    Tags for volumes are manually copied across as they may differ for each volume.
    """
    if overrides.encrypt_root_volume or overrides.enable_hibernation:
        ami = util.ami.encrypt_root_volume(ami, overrides.kms_key_id)

    if overrides.root_volume_size:
        ami = util.ami.increase_root_volume(ami, set=overrides.root_volume_size)

    elif overrides.enable_hibernation:
        ram_size = ri.RAM_SIZE[instance.size]
        ami = util.ami.increase_root_volume(ami, increase=max(ram_size, 4))

    tag_spec = util.get_tag_specifications(instance.tags, "instance")

    nics = util.instance.get_nics(instance, overrides.private_ip)

    instance.metadata_options.pop("State")
    kwargs = {
        # Without ovderrides
        "MaxCount": 1,
        "MinCount": 1,
        "TagSpecifications": tag_spec,
        "ImageId": ami.image_id,
        "KeyName": instance.key_name,
        "Placement": instance.placement,
        "MetadataOptions": instance.metadata_options,
        "EnclaveOptions": instance.enclave_options,
        "MaintenanceOptions": instance.maintenance_options,
        "CapacityReservationSpecification": instance.capacity_reservation_specification,
        "Monitoring": {
            "Enabled": True if instance.monitoring["State"] == "enabled" else False
        },
        "EbsOptimized": instance.ebs_optimized,
        "PrivateDnsNameOptions": instance.private_dns_name_options,
        "InstanceInitiatedShutdownBehavior": instance._i.describe_attribute(
            Attribute="instanceInitiatedShutdownBehavior"
        )["InstanceInitiatedShutdownBehavior"]["Value"],
        "DisableApiTermination": instance._i.describe_attribute(
            Attribute="disableApiTermination"
        )["DisableApiTermination"]["Value"],
        "InstanceInitiatedShutdownBehavior": instance._i.describe_attribute(
            Attribute="instanceInitiatedShutdownBehavior"
        )["InstanceInitiatedShutdownBehavior"]["Value"],
        # With overrides
        "BlockDeviceMappings": ami.ebs_mappings,
        "HibernationOptions": {
            "Configured": overrides.enable_hibernation
            if overrides.enable_hibernation
            else instance.hibernation_options["Configured"],
        },
        "InstanceType": overrides.instance_type
        if overrides.instance_type
        else instance.instance_type,
        "NetworkInterfaces": nics,
    }
    if instance.licenses:
        kwargs["LicenseSpecifications"] = instance.licenses

    cpu_credit_spec = util.instance.get_credit_specification(instance, client)
    if cpu_credit_spec:
        kwargs["CreditSpecification"] = cpu_credit_spec

    market_options = util.instance.get_instance_market_options(instance, client)
    if market_options:
        kwargs["InstanceMarketOptions"] = market_options

    cpu_options = util.instance.get_cpu_options(instance, client)
    if cpu_options:
        kwargs["CpuOptions"] = cpu_options

    elastic_gpus = util.instance.get_elastic_gpu_specifications(instance, client)
    if elastic_gpus:
        kwargs["ElasticGpuSpecifications"] = elastic_gpus

    inference_accelerators = util.instance.get_elastic_inference_specifications(
        instance, inference_client
    )
    if inference_accelerators:
        kwargs["ElasticInferenceAccelerators"] = inference_accelerators

    instance_profile = util.instance.get_instance_profile(instance)
    if instance_profile:
        kwargs["IamInstanceProfile"] = instance_profile

    return kwargs


def load(path=ri.SETTINGS_FILE):
    """
    Load instance settings.
    """
    log = logging.getLogger(__name__)
    try:
        with open(path, "r") as f:
            return json.load(f)
    except Exception as e:
        log.error(f"Error loading settings: {e}")
        return False


def dump(settings, path=ri.SETTINGS_FILE):
    """
    Dump instance settings.
    """

    log = logging.getLogger(__name__)
    log.info(f"Saving settings to: {path}")
    try:
        with open(path, "w") as f:
            json.dump(settings, f, indent=2)
    except Exception as e:
        log.error(f"Error saving settings: {e}")
        return False

    return True
